import json
import time
import warnings
from pathlib import Path
from typing import List

import pytest
from pymemri.data.schema import Account, Tweet
from pymemri.pod.client import PodClient
from pymemri.pod.graphql_utils import GQLQuery

from twitter_topic_model.model import TopicModelCluster
from twitter_topic_model.plugin import TwitterTopicModelPlugin
from twitter_topic_model.postprocessing import (
    filter_min_accounts_per_cluster,
    get_unique_cluster_account_ids,
)
from twitter_topic_model.schema import Cluster, ClusterEntry, TwitterTopicModel
from twitter_topic_model.utils import get_clustered_tweet_ids, get_tweets

ASSETS_DIR = Path(__file__).parent / "test_assets"
TWEET_FILE = ASSETS_DIR / "tweets.json"
TOPIC_QUERY_FILE = ASSETS_DIR / "topic_model_query.gql"


@pytest.fixture
def tweets() -> List[Tweet]:
    with open(TWEET_FILE, "r") as f:
        tweets_json = json.load(f)
    return [Tweet.from_json(t) for t in tweets_json]


@pytest.fixture
def topic_model_query() -> GQLQuery:
    with open(TOPIC_QUERY_FILE, "r") as f:
        query_str = f.read()
    return GQLQuery(query_str)


def test_get_tweets(tweets: List[Tweet]) -> None:
    client = PodClient()
    client.add_to_schema(Tweet)
    client.bulk_action(create_items=tweets)

    start_date = tweets[300].postDate
    num_tweets = len([tweet for tweet in tweets if tweet.postDate > start_date])
    tweets_from_pod = get_tweets(client, start_date=start_date)
    assert len(tweets_from_pod) == num_tweets

    tweets_from_pod = get_tweets(client, limit=100)
    assert len(tweets_from_pod) == 100


def test_fit_model_on_tweets(tweets: List[Tweet]) -> None:
    client = PodClient()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plugin = TwitterTopicModelPlugin(client=client)

    tweets = tweets[:250]
    clusters, unassigned_tweets = plugin.fit_topic_model(
        tweets, prediction_threshold=0.6
    )
    # for cluster in clusters:
    #     print(cluster.coarse_label, cluster.fine_label, max(cluster.tweet_relevance))

    for cluster in clusters:
        print(cluster.coarse_label, cluster.fine_label, len(cluster.tweets))
        # for tweet in cluster.tweets[:3]:
        #     print("\t", tweet.message)
        #     print("-------")


def test_fit_model_on_2_tweets(tweets: List[Tweet]) -> None:
    client = PodClient()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plugin = TwitterTopicModelPlugin(client=client)

    tweets = tweets[:2]
    clusters, unassigned_tweets = plugin.fit_topic_model(
        tweets, prediction_threshold=0.1
    )
    assert clusters


def test_fit_model_on_pod(tweets: List[Tweet], topic_model_query: GQLQuery) -> None:
    # Client with Tweets
    client = PodClient()
    client.add_to_schema(Tweet, Cluster, ClusterEntry, TwitterTopicModel)
    tweets = tweets[:2]
    client.bulk_action(create_items=tweets)

    # Load Plugin and fit model
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plugin = TwitterTopicModelPlugin(client=client)
    topic_model_id = plugin.create_topic_model()

    # Reset client cache and search for created topic
    client.reset_local_db()
    topic_model_item: TwitterTopicModel = client.search_graphql(
        topic_model_query, {"id": topic_model_id}
    )[0]

    # Check if all clusters are defined
    for cluster in topic_model_item.cluster:
        assert cluster.isUnassigned == True or len(cluster.clusterDescription)
        assert cluster.isUnassigned == True or cluster.category is not None
        assert cluster.isUnassigned is not None
        assert cluster.isRead == False
        for entry in cluster.entry:
            assert cluster.isUnassigned == True or entry.relevance is not None
            assert isinstance(entry.data[0], Tweet)


def test_filter_read(tweets: List[Tweet]) -> None:
    tweets = tweets[:50]

    client = PodClient()
    client.add_to_schema(Tweet, Cluster, ClusterEntry, TwitterTopicModel)

    # Create existing cluster
    create_items = []
    create_edges = []
    cluster = Cluster(isRead=True)
    create_items.append(cluster)
    for tweet in tweets:
        entry = ClusterEntry()
        create_items.append(entry)
        create_items.append(tweet)
        create_edges.append(cluster.add_edge("entry", entry))
        create_edges.append(entry.add_edge("data", tweet))
    client.bulk_action(create_items=create_items, create_edges=create_edges)

    read_ids = get_clustered_tweet_ids(client, only_read=True)
    assert len(read_ids) == len(tweets)
    all_ids = get_clustered_tweet_ids(client, only_read=False)
    assert len(all_ids) == len(tweets)

    cluster.isRead = False
    client.update_item(cluster)
    read_ids = get_clustered_tweet_ids(client, only_read=True)
    assert len(read_ids) == 0
    all_ids = get_clustered_tweet_ids(client, only_read=False)
    assert len(all_ids) == len(tweets)


def test_filter_min_accounts(tweets: List[Tweet]):
    # Use subset for speed
    tweets = tweets[:50]

    client = PodClient()
    client.add_to_schema(Tweet, Cluster, ClusterEntry, TwitterTopicModel)

    # Create existing cluster with 2 authors
    create_items = []
    create_edges = []
    cluster = Cluster(isRead=True)
    author_1 = Account(handle="Alice")
    author_2 = Account(handle="Bob")
    create_items.extend([cluster, author_1, author_2])

    for i, tweet in enumerate(tweets):
        entry = ClusterEntry()
        create_items.append(entry)
        create_items.append(tweet)
        create_edges.append(cluster.add_edge("entry", entry))
        create_edges.append(entry.add_edge("data", tweet))

        # Add author, all even tweets get author 1, else author 2
        if i % 2 == 0:
            create_edges.append(tweet.add_edge("author", author_1))
        else:
            create_edges.append(tweet.add_edge("author", author_2))
    client.bulk_action(create_items=create_items, create_edges=create_edges)

    # Use internal cluster representation
    cluster_internal = TopicModelCluster(
        coarse_label="",
        fine_label="",
        tweets=[entry.data[0] for entry in cluster.entry],
        tweet_relevance=[0.0] * len(cluster.entry),
    )
    assert len(get_unique_cluster_account_ids(cluster_internal)) == 2

    clusters = [cluster_internal]
    filtered = filter_min_accounts_per_cluster(clusters, 2)
    assert len(filtered) == 1

    filtered = filter_min_accounts_per_cluster(clusters, 3)
    assert len(filtered) == 0


@pytest.mark.skip()
def test_with_timer(tweets: List[Tweet]) -> None:
    client = PodClient()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plugin = TwitterTopicModelPlugin(client=client)

    t0 = time.time()
    clusters, unassigned_tweets = plugin.fit_topic_model(tweets)
    print("1st run {:.2f}".format(time.time() - t0))

    t0 = time.time()
    clusters, unassigned_tweets = plugin.fit_topic_model(tweets)
    print("2nd run {:.2f}".format(time.time() - t0))
