from typing import Any, Iterable, List

import numpy as np
import torch
from torch import nn


def batch_list(list_: Iterable[Any], batch_size: int):
    for i in range(0, len(list_), batch_size):
        yield list_[i : i + batch_size]


class TweetTopicClassifier(nn.Module):
    def __init__(
        self,
        input_size,
        output_size,
        num_layers=2,
        hidden_size=256,
        p_dropout=0.5,
        id2label=None,
        device="cpu",
        **kwargs
    ):
        super().__init__()
        layers = []
        prev_size = input_size
        for _ in range(num_layers - 1):
            layers.append(nn.Dropout(p=p_dropout))
            layers.append(nn.Linear(prev_size, hidden_size))
            layers.append(nn.ReLU())
            prev_size = hidden_size
        layers.append(nn.Dropout(p=p_dropout))
        layers.append(nn.Linear(prev_size, output_size))

        self.layers = nn.Sequential(*layers)
        self.id_to_label = id2label
        self.device_ = device

    def forward(self, x):
        return self.layers(x)

    def predict(self, embeddings: np.ndarray, batch_size=128, pred_threshold: float = 0.5) -> List[List[str]]:
        """Makes a multi-label prediction for a  list of embeddings."""
        res = []
        for batch in batch_list(embeddings, batch_size):
            emb_tensor = torch.FloatTensor(batch).to(self.device_)
            logits = self.forward(emb_tensor)
            pred_1h = (torch.sigmoid(logits) > pred_threshold).long()
            pred_idx = [np.where(p.cpu() == 1)[0] for p in pred_1h]
            res.extend([[self.id_to_label[p] for p in pred] for pred in pred_idx])
        return res


def load_classifier_head(model_path, device="cpu"):
    chkpt = torch.load(model_path, map_location="cpu")

    hparams = chkpt["hyper_parameters"]
    state_dict = chkpt["state_dict"]
    # training script append `classifier.*` to all params, removed here.
    state_dict = {k.split(".", 1)[1]: v for k, v in state_dict.items()}
    model = TweetTopicClassifier(**hparams, device=device)
    model.load_state_dict(state_dict, strict=True)
    model = model.to(device).eval()
    return model
