from typing import List

import gspread
import torch
from torch.utils.data import Dataset

TRAIN_LABEL_COL = 'label_hashtag'
PREDICTED_LABEL_COL = 'label_ada002'
VAL_LABEL_COL = 'label_human_1'
CONTENT_COL = 'content'

def label_is_filled(entry) -> bool:
    return entry==0 or entry==1

def is_train_data(record: dict) -> bool:
    return label_is_filled(record[TRAIN_LABEL_COL]) 

def is_val_data(record: dict) -> bool:
    return bool(
        label_is_filled(record[PREDICTED_LABEL_COL])
        and label_is_filled(record[VAL_LABEL_COL])
    )

def parse_train_record(record, label_name):
    return {
        'id': record['id'],
        'content': record[CONTENT_COL],
        'label': {
            label_name: record[TRAIN_LABEL_COL]
        }
    }

def parse_val_record(record, label_name):
    return {
        'id': record['id'],
        'content': record[CONTENT_COL],
        'label': {
            label_name: record[VAL_LABEL_COL]
        }
    }    

def load_sheet():
    gc = gspread.service_account()
    return gc.open("Twitter dataset")

def get_data_from_gsheet(sh):
    topic_sheets = [x for x in sh.worksheets() if x.title.startswith("topic")]

    # dict[uid, data] for merging tweets with the same ID
    train_data = {}
    val_data = {}
    idx2topic = {}

    for i, sheet in enumerate(topic_sheets):
        label_name = sheet.title
        idx2topic[i] = label_name

        records = sheet.get_all_records()
        for record in records:
            uid = record['id']

            if is_train_data(record):
                if uid in train_data:
                    train_data[uid]['label'][label_name] = record[TRAIN_LABEL_COL]
                else: 
                    train_data[uid] = parse_train_record(record, label_name)

            elif is_val_data(record):
                if uid in val_data:
                    val_data[uid]['label'][label_name] = record[VAL_LABEL_COL]
                else:
                    val_data[uid] = parse_val_record(record, label_name)

    train_data = list(train_data.values())
    val_data = list(val_data.values())
    return train_data, val_data, idx2topic

class TwitterTopicDataset(Dataset):
    def __init__(self, data, id2label):
        self.data = data
        self.id2label = id2label
        self.label2id = {v: k for k, v in self.id2label.items()}
        
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, i: int):
        entry = self.data[i]
        return {
            "content": entry["content"],
            "label": {self.label2id[label]: value for label, value in entry["label"].items()},
            "embedding": entry.get("embedding", None)
        }

def load_dataset():
    sh = load_sheet()
    train_data, val_data, idx2topic = get_data_from_gsheet(sh)
    
    train_dataset = TwitterTopicDataset(train_data, idx2topic)
    val_dataset = TwitterTopicDataset(val_data, idx2topic)
    return train_dataset, val_dataset

    
def embed_dataset(dataset, embedder):
    content = [item['content'] for item in dataset.data]
    embeddings = embedder.encode(content)
    for i, embedding in enumerate(embeddings):
        dataset.data[i]["embedding"] = embedding
    return dataset
