import math
import os
from pathlib import Path
from typing import Any, Dict

import evaluate
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger
from sentence_transformers import SentenceTransformer
from torch.utils.data import DataLoader

import wandb

from .dataset import embed_dataset, load_dataset
from .utils import to_multihot

PROJECT_NAME = "memri-tweettopic-2"
DEFAULT_CONFIG = {
    'lr': 2e-3,
    'p_dropout': 0.5,
    'hidden_size': 256,
    'num_layers': 2,
    'batch_size': 256,
    'embedding_model_name': 'sentence-transformers/paraphrase-multilingual-mpnet-base-v2',
}

def masked_bce_with_logits(logits, labels, mask):
    loss = F.binary_cross_entropy_with_logits(logits, labels, reduction="none")
    # Sum over unmasked labels, average over batch
    loss = (loss * mask).sum(1)
    return loss.mean(0)

def get_metrics():
    # From https://huggingface.co/datasets/cardiffnlp/tweet_topic_multi/blob/main/lm_finetuning.py
    metric_accuracy = evaluate.load("accuracy", "multilabel")
    metric_f1 = evaluate.load("f1", "multilabel")

    # metric_f1.compute(predictions=[[0, 1, 1], [1, 1, 0]], references=[[0, 1, 1], [0, 1, 0]], average='micro')
    # metric_accuracy.compute(predictions=[[0, 1, 1], [1, 1, 0]], references=[[0, 1, 1], [0, 1, 0]])

    def sigmoid(x):
        return 1 / (1 + math.exp(-x))

    def compute_metric_search(eval_pred):
        logits, labels = eval_pred
        predictions = np.array([[int(sigmoid(j) > 0.5) for j in i] for i in logits])
        return metric_f1.compute(predictions=predictions, references=labels, average='micro')

    def compute_metric_all(eval_pred):
        logits, labels = eval_pred
        predictions = np.array([[int(sigmoid(j) > 0.5) for j in i] for i in logits])
        return {
            'f1': metric_f1.compute(predictions=predictions, references=labels, average='micro')['f1'],
            'f1_macro': metric_f1.compute(predictions=predictions, references=labels, average='macro')['f1'],
            'accuracy': metric_accuracy.compute(predictions=predictions, references=labels)['accuracy']
        }
    return compute_metric_search, compute_metric_all


class TweetTopicClassifier(nn.Module):
    def __init__(self, input_size, output_size, num_layers=2, hidden_size=256, p_dropout=0.5, **kwargs):
        super().__init__()
        layers = []
        prev_size = input_size
        for _ in range(num_layers - 1):
            layers.append(nn.Dropout(p=p_dropout))
            layers.append(nn.Linear(prev_size, hidden_size))
            layers.append(nn.ReLU())
            prev_size = hidden_size
        layers.append(nn.Dropout(p=p_dropout))
        layers.append(nn.Linear(prev_size, output_size))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)


class LightningClassifier(pl.LightningModule):
    def __init__(
        self, input_size, output_size, num_layers=2, hidden_size=256, p_dropout=0.5, lr=0.01
    ):
        super().__init__()
        self.save_hyperparameters()
        self.classifier = TweetTopicClassifier(
            input_size=input_size,
            output_size=output_size,
            num_layers=num_layers,
            hidden_size=hidden_size,
            p_dropout=p_dropout,
        )
        self.lr = lr

        self.search_metrics_fn, self.metrics_fn = get_metrics()

    def calculate_metrics(self, logits, labels):
        return self.metrics_fn((logits, labels))

    def multilabel_accuracy(self, logits, label):
        probs = torch.sigmoid(logits)
        pred = probs.round()
        return (pred == label).float().mean()

    def training_step(self, batch, batch_idx):
        bx = batch["embedding"]
        by = batch["label"]
        mask = batch["label_mask"]
        out = self.classifier(bx)
        # loss = masked_bce_with_logits(out, by, mask)

        loss = F.binary_cross_entropy_with_logits(out, by)
        
        # metric_scores = self.calculate_metrics(out.detach().cpu().numpy(), by.cpu().numpy())
        metric_scores = {}
        metric_scores["loss"] = loss.detach().cpu()
        metric_scores = {"train_"+k: v for k, v in metric_scores.items()}
        self.log_dict(metric_scores)
        return loss

    def validation_step(self, batch, batch_idx):
        bx = batch["embedding"]
        by = batch["label"]
        mask = batch["label_mask"]
        out = self.classifier(bx)
        # loss = masked_bce_with_logits(out, by, mask)

        loss = F.binary_cross_entropy_with_logits(out, by)
        return {
            "loss": loss.detach().cpu().item(),
            "logits": out.detach().cpu().numpy(),
            "labels": by.cpu().numpy(),
        }

    def validation_epoch_end(self, outputs):
        logits = np.vstack([output['logits'] for output in outputs])
        labels = np.vstack([output['labels'] for output in outputs])
        loss = np.array([output['loss'] for output in outputs]).mean()

        log_dict = self.calculate_metrics(logits, labels)
        log_dict["loss"] = loss
        log_dict = {"val_"+k: v for k, v in log_dict.items()}
        self.log_dict(log_dict)

    def configure_optimizers(self):
        return torch.optim.AdamW(self.parameters(), lr=self.lr)


def collate_fn(batch, num_labels):
    """
    Return content, multihot labels, and label mask
    
    label_mask == 1 if the item has an annotation for the label, else 0
    """
    embeddings = torch.FloatTensor(np.vstack([item["embedding"] for item in batch]))
    labels = [item["label"] for item in batch]
    
    annotated_labels = [list(label.keys()) for label in labels]
    true_labels = [[k for k,v in label.items() if v==1] for label in labels]
    
    labels_1h = torch.FloatTensor(to_multihot(true_labels, num_labels))
    label_mask = torch.FloatTensor(to_multihot(annotated_labels, num_labels))
    return {"embedding": embeddings, "label": labels_1h, "label_mask": label_mask}


def train(config: Dict[str, Any]):
    wandb.init(config=config, project=PROJECT_NAME)

    train_set, valid_set = load_dataset()
    label2id = train_set.label2id
    embedder = SentenceTransformer(config['embedding_model_name'])
    train_set = embed_dataset(train_set, embedder)
    valid_set = embed_dataset(valid_set, embedder)

    collate_wrapped = lambda batch: collate_fn(batch, len(label2id))
    train_loader = DataLoader(
        train_set, batch_size=config['batch_size'], collate_fn=collate_wrapped, shuffle=True, num_workers=1)
    valid_loader = DataLoader(
        valid_set, batch_size=config['batch_size'], collate_fn=collate_wrapped, shuffle=False, num_workers=1)

    embedding_dim = embedder.get_sentence_embedding_dimension()
    num_labels = len(label2id)
    lightning_model = LightningClassifier(
        input_size=embedding_dim,
        output_size=num_labels,
        num_layers=config['num_layers'],
        hidden_size=config['hidden_size'],
        p_dropout=config['p_dropout'],
        lr=config['lr'],
    )

    wandb_logger = WandbLogger()
    project_dir = Path("./checkpoints/multi") / PROJECT_NAME / wandb_logger.experiment.id
    
    checkpoint_callback = ModelCheckpoint(
        dirpath=str(project_dir), save_top_k=1, monitor="val_f1", mode="max",
    )
    earlystopping_callback = EarlyStopping(monitor='val_f1', mode='max', patience=5)

    trainer = pl.Trainer(
        max_epochs=25,
        logger=wandb_logger,
        log_every_n_steps=1,
        accelerator="gpu",
        callbacks=[
            checkpoint_callback,
            earlystopping_callback,
        ],
    )

    trainer.fit(
        lightning_model, train_dataloaders=train_loader, val_dataloaders=valid_loader
    )

    print(f"saved best model to {checkpoint_callback.best_model_path}")

if __name__=="__main__":
    disable_wandb = False
    if disable_wandb:
        os.environ.setdefault("WANDB_MODE", "disabled")

    train(DEFAULT_CONFIG)
