from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
import torch
from pymemri.data.schema import Tweet
from sentence_transformers import SentenceTransformer

from twitter_topic_model.coarse_labels import COARSE_CATEGORY_MAPPING

from .classifier_fine_head import load_classifier_head
from .preprocessing import PreprocessedTweet

EMBEDDING_MODEL_NAME = "sentence-transformers/paraphrase-multilingual-mpnet-base-v2"


@dataclass
class TopicModelCluster:
    coarse_label: Optional[str]
    fine_label: str
    tweets: List[Union[PreprocessedTweet, Tweet]]
    tweet_relevance: Optional[List[float]]


def get_device() -> str:
    if torch.cuda.is_available():
        return "cuda"
    elif torch.has_mps:
        return "mps"
    else:
        return "cpu"


class TopicModel:
    def __init__(self) -> None:
        self.device = get_device()
        self.embedding_model = SentenceTransformer(
            EMBEDDING_MODEL_NAME, device=self.device
        )
        self.classifier_fine_head = load_classifier_head(
            "assets/i9hajgp7.ckpt",
            self.device,
        )

    def _classify_supervised(self, embeddings: np.ndarray, prediction_threshold: float = 0.5) -> List[List[str]]:
        return self.classifier_fine_head.predict(embeddings, pred_threshold=prediction_threshold)

    @staticmethod
    def _format_label(topic: Union[str, List[str]]) -> Union[str, List[str]]:
        if topic.startswith("topic_"):
            topic = topic[6:]
        return topic.replace("_", " ").title().strip()

    def _group_by_label(
        self, items: List[PreprocessedTweet], labels: List[List[str]]
    ) -> Dict[str, List[str]]:
        items_by_label = dict()
        for item, label in zip(items, labels):
            for l in label:
                if l not in items_by_label:
                    items_by_label[l] = []
                items_by_label[l].append(item)
        return items_by_label

    def _embed_tweets(self, items: List[PreprocessedTweet]) -> np.ndarray:
        sentences = [item.message_preprocessed for item in items]
        return self.embedding_model.encode(sentences)

    def fit(
        self,
        tweets: List[PreprocessedTweet],
        prediction_threshold: float = 0.5,
    ) -> List[TopicModelCluster]:

        # Generate all embeddings
        embeddings = self._embed_tweets(tweets)

        # Supervised
        supervised_labels = self._classify_supervised(embeddings, prediction_threshold)
        tweets_by_label = self._group_by_label(tweets, supervised_labels)

        clusters = [
            TopicModelCluster(
                coarse_label=COARSE_CATEGORY_MAPPING.get(k, None),
                fine_label=self._format_label(k),
                tweets=v,
                tweet_relevance=[1.0] * len(v),  # Temp hack for backwards compatibility
            )
            for k, v in tweets_by_label.items()
            if len(v) > 0
        ]

        unassigned_tweets = self.get_unassigned_tweets(tweets, clusters)
        return clusters, unassigned_tweets

    @staticmethod
    def get_unassigned_tweets(
        tweets: List[PreprocessedTweet], clusters: List[TopicModelCluster]
    ) -> List[PreprocessedTweet]:
        assigned_ids = set()
        for cluster in clusters:
            cluster_ids = [tweet.id for tweet in cluster.tweets]
            assigned_ids.update(cluster_ids)

        return [tweet for tweet in tweets if tweet.id not in assigned_ids]

    @staticmethod
    def get_embeddings_for_tweets(
        tweets: List[PreprocessedTweet], all_embeddings: np.ndarray
    ) -> np.ndarray:
        return np.take(
            all_embeddings,
            indices=[tweet.idx_ for tweet in tweets],
            axis=0,
        )

    def get_tweets_per_cluster(
        self, tweets: List[PreprocessedTweet], assignments: List[int]
    ):
        tweets_clustered = dict()
        for tweet, assignment in zip(tweets, assignments):
            if assignment not in tweets_clustered:
                tweets_clustered[assignment] = [tweet]
            else:
                tweets_clustered[assignment].append(tweet)
        return tweets_clustered

    def embed_topic_description(self, unsupervised_labels: List[str]) -> np.ndarray:
        query = ", ".join(unsupervised_labels)
        return self.embedding_model.encode(query)

    @staticmethod
    def _concat_tweets(tweets: List[PreprocessedTweet]) -> str:
        return "\n".join([tweet.message_preprocessed for tweet in tweets])
