import json
import time
from datetime import datetime
from os import environ
from typing import List, Optional, Tuple

from fastapi import HTTPException
from loguru import logger
from pymemri.data.schema import Tweet
from pymemri.plugin.constants import POD_FULL_ADDRESS_ENV
from pymemri.plugin.pluginbase import PluginBase
from pymemri.plugin.states import RUN_DAEMON
from pymemri.pod.api import DEFAULT_POD_ADDRESS
from pymemri.pod.client import PodClient
from pymemri.webserver.public_api import register_endpoint

from .model import TopicModel, TopicModelCluster
from .postprocessing import filter_min_accounts_per_cluster, filter_read_tweets
from .preprocessing import PreprocessedTweet, preprocess_tweets
from .schema import Cluster, ClusterEntry, TwitterTopicModel
from .utils import get_tweets


class TwitterTopicModelPlugin(PluginBase):
    schema_classes = [Cluster, ClusterEntry, TwitterTopicModel, Tweet]

    def __init__(
        self,
        pluginRun=None,
        client=None,
        min_cluster_tweets=2,
        run_on_start=False,
        **kwargs,
    ):
        logger.info("Setting up...")
        t0 = time.time()
        super().__init__(pluginRun, client=client, **kwargs)
        self.model = TopicModel()
        self.min_cluster_tweets = min_cluster_tweets
        self.run_on_start = run_on_start
        logger.debug(f"Setup completed in {(time.time()-t0):.2f} seconds")

    def run(self):
        if self.run_on_start:
            self.create_topic_model(max_num_tweets=500)
        self.daemon = True
        self.set_run_status(RUN_DAEMON)
        logger.info("Daemon mode started")

    def fit_topic_model(self, tweets: List[Tweet], prediction_threshold: float = 0.5):
        t0 = time.time()
        preprocessed_tweets = preprocess_tweets(tweets)

        t1 = time.time()
        logger.debug(f"Preprocessed {len(tweets)} tweets in {(t1-t0):.2f} seconds")

        clusters, unassigned_tweets = self.model.fit(
            preprocessed_tweets,
            prediction_threshold=prediction_threshold,
        )

        t2 = time.time()
        logger.debug(f"Fit topic model in {(t2-t1):.2f} seconds")

        clusters, unassigned_tweets = self._replace_preprocessed_tweets(
            clusters, unassigned_tweets, tweets
        )
        return clusters, unassigned_tweets

    def _replace_preprocessed_tweets(
        self,
        clusters: List[TopicModelCluster],
        unassigned_tweets: List[PreprocessedTweet],
        tweets: List[Tweet],
    ) -> Tuple[List[TopicModelCluster], List[Tweet]]:
        """Replaces PreprocessedTweets with Tweets in clusters and unassigned_tweets"""

        tweets_by_id = {tweet.id: tweet for tweet in tweets}

        for cluster in clusters:
            cluster.tweets = [tweets_by_id[t.id] for t in cluster.tweets]

        unassigned_tweets = [tweets_by_id[t.id] for t in unassigned_tweets]
        return clusters, unassigned_tweets

    def sync_to_pod(
        self,
        clusters: List[TopicModelCluster],
        unassigned_tweets: List[Tweet],
        start_date: Optional[datetime],
        client: Optional[PodClient] = None,
    ):
        if client is None:
            client = self.client
        topic_model_item = TwitterTopicModel(startDate=start_date)

        create_items = [topic_model_item]
        create_edges = []

        for cluster in clusters:
            cluster_item = Cluster(
                clusterDescription=cluster.fine_label,
                category=cluster.coarse_label,
                isUnassigned=False,
                isRead=False,
            )
            create_items.append(cluster_item)
            create_edges.append(topic_model_item.add_edge("cluster", cluster_item))

            for tweet, relevance in zip(cluster.tweets, cluster.tweet_relevance):
                entry = ClusterEntry(relevance=relevance)
                create_items.append(entry)
                create_edges.append(cluster_item.add_edge("entry", entry))
                create_edges.append(entry.add_edge("data", tweet))

        # Create unassigned cluster
        if len(unassigned_tweets):
            unassigned_cluster = Cluster(isUnassigned=True, isRead=False)
            create_items.append(unassigned_cluster)
            create_edges.append(
                topic_model_item.add_edge("cluster", unassigned_cluster)
            )
            for tweet in unassigned_tweets:
                entry = ClusterEntry()
                create_items.append(entry)
                create_edges.append(unassigned_cluster.add_edge("entry", entry))
                create_edges.append(entry.add_edge("data", tweet))

        client.bulk_action(
            create_items=create_items,
            create_edges=create_edges,
        )

        return topic_model_item.id

    def get_client_from_kwargs(self, kwargs):
        if kwargs is not None:
            if "url" not in kwargs:
                kwargs["url"] = environ.get(POD_FULL_ADDRESS_ENV, DEFAULT_POD_ADDRESS)

            return PodClient(create_account=False, **kwargs)
        else:
            return self.client

    @register_endpoint("/v1/create_topic_model", method="POST")
    def create_topic_model(
        self,
        start_date: Optional[datetime] = None,
        max_num_tweets: int = 500,
        pod_client_kwargs: Optional[dict] = None,
        exclude_read_tweets: bool = True,
        min_accounts_per_cluster: int = 2,
        prediction_threshold: float = 0.5,
    ) -> str:
        """Creates a topic model on the all Tweets in the Pod since `start_date`

        Args:
            start_date (Optional[datetime], optional): Only cluster tweets with postDate > start_date. Defaults to None.
            max_num_tweets (int, optional): Only cluster the last `max_num_tweets` tweet, sorted by postDate. Defaults to 500.
            pod_client_kwargs (Optional[dict], optional): Pod Credentials when running as shared plugin. Defaults to None.
            Note, you can specify POD URL in such case result will be stored address provided.
            exclude_read_tweets (bool, optional): If True, all tweets that are in previously generated clusters
                with `isRead == True` are not included in the new clusters. Defaults to True.
            min_accounts_per_cluster (int, optional): Minimum number of tweet authors per cluster.
                Cluster with less authors will not be created in the Pod. Defaults to 2.
            prediction_threshold (float, optional): Threshold between 0 and 1 for when to classify a tweet to a cluster.
                Higher is more precision, less recall. Defaults to 0.5.

        Raises:
            ValueError: Not enough tweets present in pod

        Returns:
            str: item id of created topic model
        """
        if not 0 < prediction_threshold < 1:
            raise HTTPException(
                status_code=400,
                detail="`prediction_threshold` should be between 0 and 1",
            )
        client = self.get_client_from_kwargs(pod_client_kwargs)
        client.add_to_schema(*self.schema_classes)

        tweets = get_tweets(client, start_date=start_date, limit=max_num_tweets)
        if len(tweets) < self.min_cluster_tweets:
            raise HTTPException(
                status_code=400,
                detail=f"Not enough tweets in pod ({len(tweets)}) to cluster",
            )
        clusters, unassigned_tweets = self.fit_topic_model(
            tweets, prediction_threshold=prediction_threshold
        )

        # Filters
        if exclude_read_tweets:
            clusters, unassigned_tweets = filter_read_tweets(
                clusters, unassigned_tweets, client
            )
        if min_accounts_per_cluster > 1:
            clusters = filter_min_accounts_per_cluster(
                clusters, min_accounts_per_cluster
            )

        logger.debug(
            f"Creating {len(clusters)} clusters with {len(unassigned_tweets)} unassigned tweets"
        )
        topic_model_id = self.sync_to_pod(
            clusters, unassigned_tweets, start_date, client
        )
        return topic_model_id
