from typing import List, Optional, Set

from loguru import logger
from pymemri.data.schema import Account, Tweet
from pymemri.pod.client import PodClient

from .model import TopicModelCluster
from .utils import get_clustered_tweet_ids


def filter_read_tweets(
    clusters: List[TopicModelCluster],
    unassigned_tweets: List[Tweet],
    client: PodClient
):
    read_tweet_ids = get_clustered_tweet_ids(client, only_read=True)

    # Filter read tweets from clusters
    clusters_filtered = []
    for cluster in clusters:
        # Get indices of tweets to keep first, we need to both filter tweets and relevances.
        unread_tweet_idx = [idx for idx, tweet in enumerate(cluster.tweets) if tweet.id not in read_tweet_ids]
        cluster.tweets = [cluster.tweets[idx] for idx in unread_tweet_idx]
        cluster.tweet_relevance = [cluster.tweet_relevance[idx] for idx in unread_tweet_idx]
        if len(cluster.tweets) > 0:
            clusters_filtered.append(cluster)

    # Filter read tweets from unassigned_tweets
    unassigned_filtered = [tweet for tweet in unassigned_tweets if tweet.id not in read_tweet_ids]

    return clusters_filtered, unassigned_filtered


def filter_min_accounts_per_cluster(clusters: List[TopicModelCluster], min_accounts_per_cluster: int):
    clusters_filtered = []
    for cluster in clusters:
        cluster_accounts = get_unique_cluster_account_ids(cluster)
        if len(cluster_accounts) >= min_accounts_per_cluster:
            clusters_filtered.append(cluster)
        logger.debug(f"{len(cluster_accounts)} accounts")
    return clusters_filtered


def get_unique_cluster_account_ids(cluster: TopicModelCluster) -> Set[Optional[str]]:
    unique_cluster_accounts = set()
    for tweet in cluster.tweets:
        if len(tweet.author) != 1:
            logger.warning(f"Skipped tweet ({tweet.id}) with abnormal number of authors: {len(tweet.author)}")
        else:
            unique_cluster_accounts.add(tweet.author[0].id)
    return unique_cluster_accounts       
