import re
from typing import Callable, Dict, List, Optional

import numpy as np
from ftlangdetect.detect import get_or_load_model
from pydantic import BaseModel
from pymemri.data.schema import Tweet


class PreprocessedTweet(BaseModel, arbitrary_types_allowed=True):
    id: str
    message: str
    message_preprocessed: Optional[str] = None
    embedding: Optional[np.ndarray] = None
    idx_: Optional[int] = None # For internal tracking after clustering

    @classmethod
    def from_tweet(
        cls, tweet: Tweet, preprocess_fn: Optional[Callable] = None
    ) -> "PreprocessedTweet":
        if preprocess_fn is not None:
            p = preprocess_fn(tweet)
        else:
            p = None
        return cls(id=tweet.id, message=tweet.message, message_preprocessed=p)


def preprocess_tweets(
    tweets: List[Tweet], min_num_words: int = 5
) -> List[PreprocessedTweet]:
    preprocessed = [
        PreprocessedTweet.from_tweet(tweet, preprocess_tweet) for tweet in tweets
    ]
    preprocessed = filter_tweets(preprocessed, min_num_words)
    for i, tweet in enumerate(preprocessed):
        tweet.idx_ = i
    return preprocessed


def preprocess_tweet(tweet: Tweet) -> str:
    return _remove_urls(tweet.message)


def _remove_urls(s: str) -> str:
    return re.sub(r"http\S+", "", s)


def filter_tweets(tweets: List[PreprocessedTweet], min_num_words: int = 5):
    tweets = filter_num_words(tweets, min_num_words)
    tweets = filter_english(tweets)
    return tweets


def filter_num_words(
    tweets: List[PreprocessedTweet], min_num_words: int = 5
) -> List[PreprocessedTweet]:
    return [
        tweet for tweet in tweets if len(tweet.message_preprocessed) >= min_num_words
    ]


def filter_english(tweets: List[PreprocessedTweet]) -> List[PreprocessedTweet]:
    sentences = [tweet.message.replace("\n", " ") for tweet in tweets]
    languages = detect_language(sentences)
    return [tweet for tweet, language in zip(tweets, languages) if language == "en"]


def detect_language(sentences: List[str]) -> List[str]:
    # Adapted from ftlangdetect.detect to work with multiple sentences
    model = get_or_load_model(low_memory=True)
    labels, _ = model.predict(sentences)
    labels = [label[0].replace("__label__", "") for label in labels]
    return labels
