from datetime import datetime, timezone
from typing import List, Optional, Set

from pymemri.data.schema import ItemBase
from pymemri.pod.client import PodClient

from .schema import Cluster


def get_tweets(
    client: PodClient, start_date: Optional[datetime] = None, limit: int = 10000
):

    if start_date is None:
        start_date = datetime.fromtimestamp(0, tz=timezone.utc)

    timestamp = ItemBase._datetime_to_timestamp(start_date)
    query = """
    query {
        Tweet (
            limit: $limit, order_desc: postDate, filter: {
                and: [
                    {postDate: {gt: $timestamp }},
                    {tweetType: {eq: "tweet"}}
                ]
            }
        ) {
            postDate
            message
            tweetType
            author {
                displayName
                handle
            }
        }
    }
    """
    return client.search_graphql(query, {"timestamp": timestamp, "limit": limit})


def get_clustered_tweet_ids(client: PodClient, only_read: bool = True) -> Set[Optional[str]]:
    """
    Returns ids from all tweets that are attached to Clusters

    if `only_read`, only id's from clusters where `cluster.isRead == True` are returned.  
    """
    
    query = """
    query {
        Cluster {
            isRead
            entry {
                data {
                    id
                    account {
                        id
                    }
                }
            }
        }
    }
    """

    clusters: List[Cluster] = client.search_graphql(query)
    tweet_ids = set()
    for cluster in clusters:
        if only_read and not cluster.isRead:
            continue

        for entry in cluster.entry:
            if len(entry.data):
                t = entry.data[0]
                tweet_ids.add(entry.data[0].id)
    return tweet_ids
